/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): KOMURA Takaaki <komura@astem.or.jp>
 *            Yasutaka Atarashi <atarashi@mm.media.kyoto-u.ac.jp>
 *            KASAMATSU Ken-ichi <kasamatu@lab1.kuis.kyoto-u.ac.jp>
 * started:   2001/06/16
 *
 * $Id: decoder_buf.c,v 1.10 2001/11/13 03:22:17 kasamatu Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "pipe_hd.h"
#include "decoder_buf.h"

#include "debug.h"

#define BUFFER_NEXT(x)	(((x)+1) % buf_num)
#define BUFFER_PREV(x)	(((x)+buf_num-1) % buf_num)

/*
  if buf is empty
    buf[head(==tail)]:buffering new frame in progress.
  if buf is not empty && not full && 
    assert(tail!=head)
    buf[tail]:locked or will be locked by user .
    buf[head]:buffering new frame in progress.
  if buf is full
    buf[tail(==head)]:locked or will be locked by user .
*/

static struct decoder_buf *dbuf;
static int      head, tail;
static int fd;
static int buf_size,buf_num;

static char*  headbuf_now;
static size_t headbuf_len; 
static int headbuf_ts;	/*==-1:wait for marker.  ==-2:accept new ts*/
static int buf_full;    /*head==tail && buf_full==1:buf is full ,else not full*/
static int user_lock;   /*==1: buf[tail] is locked by user now*/

int
decoder_buf_get_datanum(void)
{
  int sub;
  sub=user_lock? 1:0;
  if(buf_full){
    return buf_num-sub;
  }else{
    return ((head + buf_num - (tail+sub)) % buf_num);
  }
}


/* read from socket and write to buffer */
int
decoder_buf_read(void)
{
	ssize_t         io_ret;
	int             append;
	char p_hd[PIPE_HEADER_LEN];

	while(fd_can_read(fd,0)&& !buf_full ){
	  /* read headr  */
	  io_ret=fd_read_ignoreEOF(fd, p_hd, PIPE_HEADER_LEN);
	  if(io_ret!=PIPE_HEADER_LEN){
	    return 0;
	  }
	  append=1;
	  if(headbuf_ts==-1){
	    /* drop packet */
	    d3_printf("drop packet\n");
	    append=0;
	  }else{
	    /* ts and len check */
	    if(headbuf_ts==-2){
	      headbuf_ts=pipe_get_timestamp(p_hd);
	      d3_printf("start new packet seq,ts%u\n",headbuf_ts);
	    }else{
	      if(headbuf_ts!=pipe_get_timestamp(p_hd)
		 || buf_size - headbuf_len<pipe_get_length(p_hd)
		 || pipe_get_length(p_hd)==0)
		{
		d3_printf("drop packet seq \n");
		headbuf_ts=-1;
		append=0;
		}
	    }
	  }
	  /* append to buf or drop   */
	  if(append){
	    io_ret=fd_read_ignoreEOF(fd,headbuf_now,pipe_get_length(p_hd));
	    if(io_ret != pipe_get_length(p_hd)){
	      return 0;
	    }
	    headbuf_now+=io_ret;
	    headbuf_len+=io_ret;
	    if(pipe_get_marker(p_hd)==1){
	      dbuf[head].len = headbuf_len;
	      dbuf[head].ts=pipe_get_timestamp(p_hd);
	      d2_printf("buffered frame ts%u len%d\n",headbuf_ts,headbuf_len);
	      head = BUFFER_NEXT(head);
	      headbuf_len=0;
	      headbuf_now=dbuf[head].buf;
	      headbuf_ts=-2;
	      if(head==tail){
		d2_printf("buffere full h%d\n",head);
		buf_full=1;
	      }
	    }
	  }else{
	    io_ret=fd_lseek_ignoreEOF(fd,pipe_get_length(p_hd));
	    if(io_ret <0){
	      return 0;
	    }
	    if(pipe_get_marker(p_hd)==1){
	      headbuf_len=0;
	      headbuf_now=dbuf[head].buf;
	      headbuf_ts=-2;
	    }
	  }
	}
	if(buf_full==1){
	  return 2;
	}
	return 1;
}


int             decoder_buf_unlock(){
         if(user_lock==1){
	   user_lock=0;
	    tail = BUFFER_NEXT(tail);
	    buf_full=0;
            return 1;
         }
	 return 0;
}
/* read from buffer and write to user's buf */
int             decoder_buf_lock(void **ret_buf,u_int32_t *ret_ts){
	int             len;

	if (tail == head && !buf_full) {/* buffer empty */
		return 0;
	}
	len = dbuf[tail].len;
	if (len == -1) {	/* end of all data */
		return -1;
	} else if (len == 0) {
		d1_printf("\ndecoder_buf_get: data size = 0");
		return 0;
	}
	if (len > buf_size) {
		e_printf("\ndbuf[tail].len > buf_size, tail=%d, len=%d",
			 tail, len);
		exit(-1);
	}
	if(user_lock){
	  tail = BUFFER_NEXT(tail);
	  buf_full=0;
	}else{
	  user_lock=1;
	}

	*ret_ts=dbuf[tail].ts;
#if 0
	memcpy(buf, dbuf[tail].buf, len);
#else
	*ret_buf=dbuf[tail].buf;
#endif
	//d_printf("<%d>", decoder_buf_get_datanum());
	return len;
}

void decoder_buf_rm(int num){
  int add;
  if(num<=0)return;
  add=user_lock? 0:1;
  user_lock=0;
  buf_full=0;
  tail=(((tail)+num+add) % buf_num);
  return;
}

int
decoder_buf_prebuf(void)
{
	while (decoder_buf_get_datanum() < buf_num / 2) {
	  fd_can_read(STDIN_FILENO,100*1000);
	  decoder_buf_read();
	}
	return decoder_buf_get_datanum();
}


void
decoder_buf_init(int fd_,int buf_size_,int buf_num_)
{
	int             i;
	u_int8_t       *p;
	fd=fd_;
	buf_size=buf_size_;
	buf_num=buf_num_;

	d1_printf("decoder_buf_init fd=%d dsize=%d dnum=%d\n"
		  ,fd,buf_size,buf_num);

	
	dbuf = (struct decoder_buf *) malloc(sizeof(struct decoder_buf)* buf_num);
	p= malloc(buf_size*buf_num);
	if (dbuf == NULL || p== NULL) {
		d_printf("decode_buff: cannot malloc for buf\n");
		exit(-1);
	}
	for (i = 0; i < buf_num; i++) {
		dbuf[i].buf=p+i*buf_size;
	}
	head = tail = 0;
	buf_full=0;
	headbuf_now=dbuf[head].buf;
	headbuf_len=0;
	user_lock=0;
	return;
}
