/* 
 * Error-correcting Multicast on Open Nodes (EMON)
 *
 * This software is placed in public domain.
 */ 

/*
 * author(s): Yasutaka Atarashi <atarashi@mm.media.kyoto-u.ac.jp>
 * started:   2001/06/26
 *
 * $Id: jpegsave.c,v 1.1 2001/10/03 09:04:10 kasamatu Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>

#include "def_value.h"
#include "libfec.h"
#include "pipe_hd.h"
#include "debug.h"

#define FMT_MEDIA "%04d.jpg"

static struct opt_values {	/* set by getenv and/or getopt */
	char           *fmt_media;
	int             rs_N, rs_F;
	int             rs_K;	/* not identical variable */
	int             plen;
	int             num_files;
}               OPT;

int             debug_level;

int
opt_etc(int argc, char *argv[])
{
	int             ch;	/* for getopt */

	char           *temp;
	char           *opts = "m:n:L:N:F:D:";
	char           *helpmes =
	"usage : %s [-m fmt_media] [-n n] [-L packet len] [-N n] [-F n] [-D n]\n"
	" options\n"
	" <general>\n"
	"  m <str> : Media filename format (e.g. %%04d.jpg)\n"
	"  n <n>   : # of writing files\n"
	"  L <n>   : length of a element\n"
	" <Read Solomon>\n"
	"  N <n>   : # of total packets\n"
	"  F <n>   : # of RS packets\n"
	" <debug>\n"
	"  D <n>   : debug level\n"
	               ;
	/* set default value */
	OPT.fmt_media = FMT_MEDIA;
	OPT.rs_N = DEF_EMON_RS_N;
	OPT.rs_F = DEF_EMON_RS_F;
	OPT.plen = DEF_EMON_PLEN;

	/* get environment variables */

	if ((temp = getenv("EMON_PLEN")) != NULL)
		OPT.plen = atoi(temp);
	if ((temp = getenv("EMON_RS_N")) != NULL)
		OPT.rs_N = atoi(temp);
	if ((temp = getenv("EMON_RS_F")) != NULL)
		OPT.rs_F = atoi(temp);

	while ((ch = getopt(argc, argv, opts)) != -1) {
		switch (ch) {
		case 'm':
			OPT.fmt_media = optarg;
			break;
		case 'n':
			OPT.num_files = atoi(optarg);
			break;
		case 'L':
			OPT.plen = atoi(optarg);
			break;
		case 'N':	/* RS N */
			OPT.rs_N = atoi(optarg);
			break;
		case 'F':   /* RS F */
			OPT.rs_F = atoi(optarg);
			break;
		case 'D':
			debug_level = atoi(optarg);
			break;
		default:
			fprintf(stderr, helpmes, argv[0]);
			return -1;
		}
	}

	OPT.rs_K=OPT.rs_N-OPT.rs_F;
	return 0;
}


int
open_nextfile(void)
{
	static int      fileno = 0;
	static char     name[255];

	int             fd = -1;/* file descriptor */

	for (;;) {
		if(OPT.num_files && fileno>OPT.num_files) {
			e_printf("Too many files %s\n", name);
			return -1;
		}
		snprintf(name, sizeof(name), OPT.fmt_media, fileno++);
		fd = open(name, O_RDWR | O_CREAT, 0644);
		if (fd == -1) {
			e_printf("cannot open %s\n", name);
			return -1;
		} else {
			d1_printf("\npipesave: open %s", name);
			break;	/* == return fd */
		}
	}
	return fd;
}

int
main(int argc, char *argv[])
{
	char           *buffer;
	char            header[PIPE_HEADER_LEN];
	int             eras_pos[FEC_N_MAX], no_eras, fd;

	if (opt_etc(argc, argv) == -1) {
		return -1;
	}
	if (isatty(STDIN)) {
		e_printf("This program must be used with input connecting to pipe.\n");
		return -1;
	}
	if ((buffer = malloc(OPT.rs_N * OPT.plen)) == NULL) {
		e_printf("Cannot malloc for buffer.\n");
		return -1;
	}
	memset(header, 0, PIPE_HEADER_LEN);
	memset(buffer, 0, OPT.rs_N * OPT.plen);

	while(pipe_blocked_read_message_ex(STDIN, header, buffer, eras_pos, &no_eras, OPT.rs_N, OPT.plen)!=PIPE_ERROR) {
#if 0
		fprintf(stderr,"%d\n",pipe_get_length(header));
#endif
		fd=open_nextfile();
		if(fd<0) break;
		if(pipe_get_length(header) < OPT.rs_K*OPT.plen)
			write(fd, buffer, pipe_get_length(header));
		else
			write(fd, buffer, OPT.rs_K*OPT.plen);
		close(fd);
	}

	return 0;
}
